from __future__ import annotations

from functools import cached_property
from pathlib import Path

from seadex._types import File, StrPath
from seadex._utils import realpath


class SeaDexTorrent:
    def __init__(self, file: StrPath) -> None:
        """
        Class to handle torrent files for SeaDex.

        Parameters
        ----------
        file : StrPath
            The path to the torrent file.

        """
        try:
            import torf
        except ModuleNotFoundError:  # pragma: no cover
            msg = (
                "The 'torf' library is required to use the SeaDexTorrent class.\n"
                "Please ensure you have installed SeaDex with the 'torrent' extra, "
                "typically specified as 'seadex[torrent]'."
            )
            raise ModuleNotFoundError(msg) from None

        self._file = realpath(file)
        self._torrent = torf.Torrent.read(self._file)

    @property
    def file(self) -> Path:
        """Resolved path to the torrent file."""
        return self._file

    @cached_property
    def filelist(self) -> tuple[File, ...]:
        """List of files within the torrent."""
        files = []
        for file in self._torrent.files:
            files.append(File(name=Path(file).as_posix(), size=file.size))
        return tuple(files)

    def sanitize(self, *, destination: StrPath | None = None, overwrite: bool = False) -> Path:
        """
        Sanitizes the torrent file by removing sensitive data and optionally saves it to a new location.

        Parameters
        ----------
        destination : StrPath | None, optional
            The destination path to save the sanitized torrent. If None, the sanitized file is saved in place.
        overwrite : bool, optional
            If True, overwrites the existing file or destination file if it exists.

        Returns
        -------
        Path
            The path to the sanitized torrent file.

        Raises
        ------
        FileExistsError
            - If `destination` is None and `overwrite` is False.
            - If `destination` already exists and `overwrite` is False.

        Notes
        -----
        - If the torrent file is public (i.e., not marked as private), it is returned as is.
        - The following fields are removed from the torrent file if it is private:
            - Trackers
            - Web seeds
            - HTTP seeds
            - Private flag
            - Comment
            - Creation date
            - Created by field
            - Source field
        - The torrent's `infohash` is randomized.

        """

        if not self._torrent.private:
            # Public torrent
            return self.file

        self._torrent.trackers = None
        self._torrent.webseeds = None
        self._torrent.httpseeds = None
        self._torrent.private = None
        self._torrent.comment = None
        self._torrent.creation_date = None
        self._torrent.created_by = None
        self._torrent.source = None
        self._torrent.randomize_infohash = True

        if destination is None:
            if overwrite is False:
                raise FileExistsError(self.file)
            self._torrent.write(self.file, overwrite=overwrite)
            return self.file
        destination = realpath(destination)
        self._torrent.write(destination, overwrite=overwrite)
        return destination



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING
from uuid import uuid4
from zipfile import ZipFile

import pytest

from seadex import BackupFile, BadBackupFileError, SeaDexBackup

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_httpx import HTTPXMock


def test_backupfile() -> None:
    backupfile = BackupFile(
        name="test.zip",
        size=1024,
        modified_time=datetime(2024, 9, 12, 18, 14, 33, 816632, tzinfo=timezone.utc),
    )

    assert BackupFile.from_json(backupfile.to_json()) == BackupFile.from_dict(backupfile.to_dict())

    assert str(backupfile) == backupfile.name
    assert backupfile.size == 1024
    assert backupfile.modified_time == datetime(2024, 9, 12, 18, 14, 33, 816632, tzinfo=timezone.utc)


def test_backup_properties(seadex_backup: SeaDexBackup, httpx_mock: HTTPXMock) -> None:
    sample = [
        {"key": "@auto_pb_backup_sea_dex_20241116000000.zip", "size": 65571703, "modified": "2024-11-16 00:00:04.982Z"},
        {"key": "@auto_pb_backup_sea_dex_20241117000000.zip", "size": 65526931, "modified": "2024-11-17 00:00:03.423Z"},
        {"key": "@auto_pb_backup_sea_dex_20241118000000.zip", "size": 65899588, "modified": "2024-11-18 00:00:03.390Z"},
        {"key": "@auto_pb_backup_sea_dex_20241119000000.zip", "size": 66651413, "modified": "2024-11-19 00:00:03.511Z"},
        {"key": "@auto_pb_backup_sea_dex_20241120000000.zip", "size": 66249632, "modified": "2024-11-20 00:00:03.784Z"},
        {"key": "@auto_pb_backup_sea_dex_20241121000000.zip", "size": 65304056, "modified": "2024-11-21 00:00:03.661Z"},
        {"key": "@auto_pb_backup_sea_dex_20241122000000.zip", "size": 65847001, "modified": "2024-11-22 00:00:03.487Z"},
    ]
    httpx_mock.add_response(url="https://releases.moe/api/backups", json=sample, is_reusable=True)
    assert seadex_backup.base_url == "https://releases.moe"
    assert len(seadex_backup.get_backups()) == 7
    assert seadex_backup.get_latest_backup() == BackupFile(
        name="@auto_pb_backup_sea_dex_20241122000000.zip",
        size=65847001,
        modified_time=datetime(2024, 11, 22, 0, 0, 3, 487000, tzinfo=timezone.utc),
    )


def test_backup_download_with_invalid_destination(seadex_backup: SeaDexBackup, tmp_path: Path) -> None:
    with pytest.raises(NotADirectoryError):
        seadex_backup.download(destination=tmp_path / "doesnt exist")


def test_backup_download(
    seadex_backup: SeaDexBackup, httpx_mock: HTTPXMock, tmp_path_factory: pytest.TempPathFactory
) -> None:
    sample = [
        {"key": "@auto_pb_backup_sea_dex_20241120000000.zip", "size": 66249632, "modified": "2024-11-20 00:00:03.784Z"},
        {"key": "@auto_pb_backup_sea_dex_20241121000000.zip", "size": 65304056, "modified": "2024-11-21 00:00:03.661Z"},
        {"key": "@auto_pb_backup_sea_dex_20241122000000.zip", "size": 65847001, "modified": "2024-11-22 00:00:03.487Z"},
    ]
    httpx_mock.add_response(url="https://releases.moe/api/backups", json=sample, is_reusable=True)

    latest_backup = seadex_backup.get_latest_backup()
    sample_backup_file = tmp_path_factory.mktemp("blah") / "sample_backup.zip"

    with ZipFile(sample_backup_file, "w") as f:
        f.writestr("data.db", data=b"hello world")

    httpx_mock.add_response(
        url=f"https://releases.moe/api/backups/{latest_backup.name}?token=secret",
        content=sample_backup_file.read_bytes(),
        is_reusable=True,
    )

    zip_with_bad_crc = f"{uuid4()}.zip"

    httpx_mock.add_response(
        url=f"https://releases.moe/api/backups/{zip_with_bad_crc}?token=secret",
        # https://github.com/python/cpython/blob/f1e74248025b36a0c5d12f72c4ab713f4682f523/Lib/test/test_zipfile/test_core.py#L2435-L2443
        content=(
            b"PK\003\004\024\0\0\0\0\0 \213\212;:r"
            b"\253\377\f\0\0\0\f\0\0\0\005\0\0\000af"
            b"ilehello,AworldP"
            b"K\001\002\024\003\024\0\0\0\0\0 \213\212;:"
            b"r\253\377\f\0\0\0\f\0\0\0\005\0\0\0\0"
            b"\0\0\0\0\0\0\0\200\001\0\0\0\000afi"
            b"lePK\005\006\0\0\0\0\001\0\001\0003\000"
            b"\0\0/\0\0\0\0\0"
        ),
    )

    backup1 = seadex_backup.download(destination=tmp_path_factory.mktemp("blah2"))
    backup2 = seadex_backup.download(latest_backup, destination=tmp_path_factory.mktemp("blah3"))
    backup3 = seadex_backup.download(
        "@auto_pb_backup_sea_dex_20241122000000.zip", destination=tmp_path_factory.mktemp("blah3")
    )

    assert backup1.name == latest_backup.name
    assert backup2.name == latest_backup.name
    assert backup3.name == latest_backup.name

    with pytest.raises(TypeError):
        seadex_backup.download(1213123, destination=tmp_path_factory.mktemp("blah4"))  # type: ignore[arg-type]

    with pytest.raises(BadBackupFileError):
        seadex_backup.download(zip_with_bad_crc, destination=tmp_path_factory.mktemp("blah5"))


def test_backup_create(seadex_backup: SeaDexBackup) -> None:
    with pytest.raises(ValueError):
        seadex_backup.create("@invalid_character.zip")


def test_backup_create_with_invalid_filename(seadex_backup: SeaDexBackup, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://releases.moe/api/backups",
        method="POST",
        json={"name": "seadex-backup.zip"},
        headers={"Authorization": "secret"},
    )
    sample = [
        {"key": "@auto_pb_backup_sea_dex_20241120000000.zip", "size": 66249632, "modified": "2024-11-20 00:00:03.784Z"},
        {"key": "@auto_pb_backup_sea_dex_20241121000000.zip", "size": 65304056, "modified": "2024-11-21 00:00:03.661Z"},
        {"key": "seadex-backup.zip", "size": 65847001, "modified": "2024-11-23 00:00:03.487Z"},
    ]
    httpx_mock.add_response(url="https://releases.moe/api/backups", method="GET", json=sample)

    backup = seadex_backup.create("seadex-backup.zip")
    assert backup.name == "seadex-backup.zip"


def test_backup_delete(seadex_backup: SeaDexBackup, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url="https://releases.moe/api/backups/seadex-backup.zip", method="DELETE")
    seadex_backup.delete("seadex-backup.zip")



from __future__ import annotations

from datetime import datetime
from os import PathLike
from typing import TYPE_CHECKING, Any, TypeAlias
from urllib.parse import urljoin

import msgspec
from natsort import natsorted, ns

from seadex._enums import Tracker

if TYPE_CHECKING:
    from typing_extensions import Self


StrPath: TypeAlias = str | PathLike[str]
"""String or path-like objects"""


class Base(
    msgspec.Struct,
    forbid_unknown_fields=True,
    frozen=True,
    kw_only=True,
):
    """Base class for AniList data structures."""

    @classmethod
    def from_dict(cls, data: dict[str, Any], /) -> Self:
        """
        Create an instance of this class from a dictionary.

        Parameters
        ----------
        data : dict[str, Any]
            Dictionary representing the instance of this class.

        Returns
        -------
        Self
            An instance of this class.

        """
        return msgspec.convert(data, type=cls)

    def to_dict(self) -> dict[str, Any]:
        """
        Serialize the instance of this class into a dictionary.

        Returns
        -------
        dict[str, Any]
            Dictionary representing the instance of this class.

        """
        return msgspec.to_builtins(self)  # type: ignore[no-any-return]

    @classmethod
    def from_json(cls, data: str | bytes, /) -> Self:
        """
        Create an instance of this class from JSON data.

        Parameters
        ----------
        data : str | bytes
            JSON data representing the instance of this class.

        Returns
        -------
        Self
            An instance of this class.

        """
        return msgspec.json.decode(data, type=cls)

    def to_json(self, *, indent: int = 2) -> str:
        """
        Serialize the instance of this class into a JSON string.

        Parameters
        ----------
        indent : int, optional
            Number of spaces for indentation.
            Set to 0 for a single line with spacing,
            or negative to minimize size by removing extra whitespace.

        Returns
        -------
        str
            JSON string representing this class.

        """
        jsonified = msgspec.json.encode(self)
        return msgspec.json.format(jsonified, indent=indent).decode()


class File(Base, frozen=True, kw_only=True):
    """Represents a file in the torrent."""

    name: str
    """The name of the file."""
    size: int
    """The size of the file in bytes."""

    def __str__(self) -> str:
        """Implement the string representation. Equivalent to `File.name`."""
        return self.name


class TorrentRecord(Base, frozen=True, kw_only=True):
    """Represents a single torrent record within a SeaDex entry."""

    collection_id: str
    """The ID of the collection the torrent record belongs to."""
    collection_name: str
    """The name of the collection the torrent record belongs to."""
    created_at: datetime
    """The timestamp of when the torrent record was created."""
    is_dual_audio: bool
    """Whether the torrent contains both Japanese and English audio tracks."""
    files: tuple[File, ...]
    """A tuple of `File` objects representing the files in the torrent."""
    id: str
    """The ID of the torrent record."""
    infohash: str | None
    """The infohash of the torrent if available, otherwise `None` (private torrents)."""
    is_best: bool
    """Whether this torrent is marked as the "best"."""
    release_group: str
    """The name of the group that released the torrent."""
    tracker: Tracker
    """The tracker where the torrent is hosted."""
    updated_at: datetime
    """The timestamp of when the torrent record was last updated."""
    url: str
    """The URL of the torrent."""
    size: int
    """The total size of the torrent, calculated by summing the sizes of all files."""

    @classmethod
    def from_dict(cls, data: dict[str, Any], /) -> Self:
        """
        Create an instance of this class from a dictionary.

        Parameters
        ----------
        data : dict[str, Any]
            Dictionary representing the instance of this class.

        Returns
        -------
        Self
            An instance of this class.

        """
        try:
            # Attempt a strict conversion, assuming data
            # comes from TorrentRecord.to_dict()
            return msgspec.convert(data, type=cls)
        except msgspec.ValidationError:
            # Failed, let's attempt a laxer conversion,
            # assuming the data comes from the SeaDex API.
            files = set()
            size = 0

            for file in data["files"]:
                files.add(File(name=file["name"], size=file["length"]))
                size += file["length"]

            # SeaDex API uses "<redacted>" to indicate that the torrent has no infohash (because it's private).
            # This replaces it with None for a more pythonic approach.
            infohash = None if data["infoHash"] == "<redacted>" else data["infoHash"]

            # Private trackers only have a relative URL on SeaDex
            # This will transform said relative URLs into absolute URLs.
            tracker = Tracker(data["tracker"])
            url: str = data["url"]
            if not url.startswith(tracker.url) and tracker.is_private():
                url = urljoin(tracker.url, url)

            kwargs = {
                "collection_id": data["collectionId"],
                "collection_name": data["collectionName"],
                "created_at": data["created"],
                "is_dual_audio": data["dualAudio"],
                "files": natsorted(files, alg=ns.PATH),
                "id": data["id"],
                "infohash": infohash,
                "is_best": data["isBest"],
                "release_group": data["releaseGroup"],
                "tracker": tracker,
                "updated_at": data["updated"],
                "url": url,
                "size": size,
            }
            return msgspec.convert(kwargs, type=cls, strict=False)


class EntryRecord(Base, frozen=True, kw_only=True):
    """Represents a single anime entry in SeaDex."""

    anilist_id: int
    """The AniList ID of the anime."""
    collection_id: str
    """The ID of the collection the entry belongs to."""
    collection_name: str
    """The name of the collection the entry belongs to."""
    comparisons: tuple[str, ...]
    """A tuple of comparison urls."""
    created_at: datetime
    """The timestamp of when the entry was created."""
    id: str
    """The ID of the entry."""
    is_incomplete: bool
    """Whether the entry is considered incomplete."""
    notes: str
    """Additional notes about the entry."""
    theoretical_best: str | None
    """The theoretical best release for the entry, if known."""
    torrents: tuple[TorrentRecord, ...]
    """A tuple of `TorrentRecord` objects associated with the entry."""
    updated_at: datetime
    """The timestamp of when the entry was last updated."""
    url: str
    """The URL of the entry."""
    size: int
    """The total size of the entry, calculated by summing the sizes of all files in all torrents."""

    @classmethod
    def from_dict(cls, data: dict[str, Any], /) -> Self:
        """
        Create an instance of this class from a dictionary.

        Parameters
        ----------
        data : dict[str, Any]
            Dictionary representing the instance of this class.

        Returns
        -------
        Self
            An instance of this class.

        """
        try:
            # Attempt a strict conversion, assuming data
            # comes from EntryRecord.to_dict()
            return msgspec.convert(data, type=cls)
        except msgspec.ValidationError:
            # Failed, let's attempt a laxer conversion,
            # assuming the data comes from the SeaDex API.

            # Grab all the torrents in the entry
            # And caclculate the total size of the entry.
            torrents: list[TorrentRecord] = []
            size = 0

            for torrent in data["expand"]["trs"]:
                tr = TorrentRecord.from_dict(torrent)
                size += tr.size
                torrents.append(tr)

            # "comparison" is a comma seperated string,
            # so we'll transform it into a set
            comparisons = {i.strip() for i in data["comparison"].split(",") if i}

            # "theoreticalBest" is an empty string in the API
            # if there's no theoretical best, we'll replace it with None
            theoretical_best = data["theoreticalBest"] if data["theoreticalBest"] else None

            anilist_id = data["alID"]

            kwargs = {
                "anilist_id": anilist_id,
                "collection_id": data["collectionId"],
                "collection_name": data["collectionName"],
                "comparisons": comparisons,
                "created_at": data["created"],
                "id": data["id"],
                "is_incomplete": data["incomplete"],
                "notes": data["notes"],
                "theoretical_best": theoretical_best,
                "updated_at": data["updated"],
                "torrents": torrents,
                "url": f"https://releases.moe/{anilist_id}/",
                "size": size,
            }
            return msgspec.convert(kwargs, type=cls, strict=False)



from __future__ import annotations

import re
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Any
from urllib.parse import urljoin
from zipfile import ZipFile

import httpx
import msgspec
from atomicwriter import AtomicWriter
from httpx import Client

from seadex._exceptions import BadBackupFileError
from seadex._types import Base, StrPath
from seadex._utils import httpx_client, realpath

if TYPE_CHECKING:
    from typing_extensions import Self


class BackupFile(Base, frozen=True, kw_only=True):
    """Represents a backup file."""

    name: str
    """The name of the backup file."""
    size: int
    """The size of the backup file in bytes."""
    modified_time: datetime
    """The last modified time of the backup file."""

    def __str__(self) -> str:
        """Implement the string representation. Equivalent to `BackupFile.name`."""
        return self.name

    @classmethod
    def from_dict(cls, data: dict[str, Any], /) -> Self:
        """
        Create an instance of this class from a dictionary.

        Parameters
        ----------
        data : dict[str, Any]
            Dictionary representing the instance of this class.

        Returns
        -------
        Self
            An instance of this class.

        """
        try:
            # Attempt a strict conversion, assuming data
            # comes from TorrentRecord.to_dict()
            return msgspec.convert(data, type=cls)
        except msgspec.ValidationError:
            # Failed, let's attempt a laxer conversion,
            # assuming the data comes from the SeaDex API.
            kwargs = {
                "name": data["key"],
                "modified_time": data["modified"],
                "size": data["size"],
            }
            return msgspec.convert(kwargs, type=cls, strict=False)


class SeaDexBackup:
    def __init__(
        self,
        email: str,
        password: str,
        *,
        base_url: str = "https://releases.moe",
        client: Client | None = None,
    ) -> None:
        """
        Client to interact with the SeaDex backup API.

        Parameters
        ----------
        email : str
            The email address for authentication.
        password : str
            The password for authentication.
        base_url : str, optional
            The base URL of SeaDex, used for constructing API queries.
        client : Client, optional
            An [`httpx.Client`][httpx.Client] instance used to make requests to SeaDex.

            [httpx.Client]: https://www.python-httpx.org/advanced/#client

        Examples
        --------
        ```py
        with SeaDexBackup("me@email.com", "password") as seadex_backup:
            print(seadex_backup.get_latest_backup())
            #> @auto_pb_backup_sea_dex_20241122000000.zip
        ```

        Notes
        -----
        Only SeaDex admins can use this! Logging in with a non-admin account will result in failure.

        """
        self._base_url = base_url
        # Increase client timeout to 60s for backup operations.
        # Their large and growing size (>= 160 MB) can cause
        # methods like create() to exceed the default 5s timeout.
        self._client = httpx_client(timeout=httpx.Timeout(60)) if client is None else client
        self._admin_token = self._auth_with_password(email, password)

    @property
    def base_url(self) -> str:
        """
        Base URL, used for constructing API queries.
        """
        return self._base_url

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: object) -> None:
        self.close()

    def close(self) -> None:
        """
        Close the underlying HTTP client connection.
        """
        self._client.close()

    def _url_for(self, endpoint: str) -> str:
        return urljoin(self._base_url, endpoint)

    def _auth_with_password(self, email: str, password: str) -> str:
        response = self._client.post(
            self._url_for("/api/admins/auth-with-password"), json={"identity": email, "password": password}
        )
        admin = response.raise_for_status().json()
        return admin["token"]  # type: ignore[no-any-return]

    def _get_file_token(self) -> str:
        response = self._client.post(self._url_for("/api/files/token"), headers={"Authorization": self._admin_token})
        return response.raise_for_status().json()["token"]  # type: ignore[no-any-return]

    def get_backups(self) -> tuple[BackupFile, ...]:
        """
        Retrieve a tuple of backup files.

        Returns
        -------
        tuple[BackupFile, ...]
            A tuple of backup files, sorted by the modified date.

        """

        response = self._client.get(
            "https://releases.moe/api/backups", headers={"Authorization": self._admin_token}
        ).json()
        backups = [BackupFile.from_dict(backup) for backup in response]
        return tuple(sorted(backups, key=lambda f: f.modified_time))

    def get_latest_backup(self) -> BackupFile:
        """
        Retrieve the latest backup file.

        Returns
        -------
        BackupFile
            The latest backup file.

        """
        return self.get_backups()[-1]

    def download(
        self, file: str | BackupFile | None = None, *, destination: StrPath | None = None, overwrite: bool = False
    ) -> Path:
        """
        Download the specified backup file to the given destination directory.

        Parameters
        ----------
        file : str | BackupFile | None, optional
            The backup file to download.
            If `None`, downloads the [latest existing backup][seadex.SeaDexBackup.get_latest_backup].
        destination : StrPath | None, optional
            The destination directory to save the backup.
        overwrite : bool, optional
            Whether to overwrite the file if it already exists.

        Returns
        -------
        Path
            The path to the downloaded backup file.

        Raises
        ------
        NotADirectoryError
            If the destination is not a valid directory.
        BadBackupFileError
            If the downloaded backup file fails integrity check.
        TypeError
            If the provided `file` argument has an invalid type.

        """
        destination = Path.cwd() if destination is None else realpath(destination)

        if not destination.is_dir():
            errmsg = f"{destination} must be an existing directory!"
            raise NotADirectoryError(errmsg)

        match file:
            case None:
                key = self.get_latest_backup().name
            case str():
                key = file
            case BackupFile():
                key = file.name
            case _:
                errmsg = f"'file' must be a string or path-like, not {type(file).__name__}."
                raise TypeError(errmsg)

        outfile = destination / key

        with AtomicWriter(outfile, overwrite=overwrite) as f:
            url = self._url_for(f"/api/backups/{key}")
            params = {"token": self._get_file_token()}
            with self._client.stream("GET", url, params=params) as response:
                for chunk in response.iter_bytes(1024 * 1024):
                    f.write_bytes(chunk)

        with ZipFile(outfile) as archive:
            check = archive.testzip()

        if check is not None:
            outfile.unlink(missing_ok=True)
            errmsg = f"{outfile} failed integrity check!"
            raise BadBackupFileError(errmsg)

        return outfile

    def create(self, filename: str) -> BackupFile:
        """
        Create a new backup with the specified filename.

        Parameters
        ----------
        filename : str
            The name to assign to the backup file.
            The filename must contain only lowercase alphabets, numbers, hyphens, or underscores.
            It may also include formatting options as supported by [`datetime.strftime`][datetime.datetime.strftime].

        Returns
        -------
        BackupFile
            The newly created backup file.

        Raises
        ------
        ValueError
            If the filename is invalid.

        """
        _filename = filename.removesuffix(".zip") + ".zip"
        _filename = datetime.now(timezone.utc).strftime(_filename).casefold()

        if re.match(r"^([a-z0-9_-]+\.zip)$", _filename) is None:
            # The API forbids anything else, so we need to enforce it.
            errmsg = (
                f"Invalid filename: {_filename!r}. "
                "The filename may only contain lowercase alphabets, "
                "numbers, hyphens, or underscores."
            )
            raise ValueError(errmsg)

        self._client.post(
            self._url_for("/api/backups"),
            json={"name": _filename},
            headers={"Authorization": self._admin_token},
        ).raise_for_status()

        # https://boltons.readthedocs.io/en/latest/iterutils.html#boltons.iterutils.first
        return next(filter(lambda member: member.name == _filename, self.get_backups()))

    def delete(self, file: str | BackupFile) -> None:
        """
        Delete the specified backup file.

        Parameters
        ----------
        file : str | BackupFile
            The backup file to delete.

        Returns
        -------
        None

        """
        self._client.delete(
            self._url_for(f"/api/backups/{file}"), headers={"Authorization": self._admin_token}
        ).raise_for_status()



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from seadex import EntryNotFoundError, SeaDexEntry

if TYPE_CHECKING:
    from pytest_httpx import HTTPXMock


def test_entry_not_found_from_anilist_id(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?perPage=500&expand=trs&filter=alID%3D98329837198378237183918&skipTotal=true",
        json={"page": 1, "perPage": 30, "totalItems": 0, "totalPages": 0, "items": []},
    )

    with pytest.raises(EntryNotFoundError):
        seadex_entry.from_id(98329837198378237183918)


def test_entry_not_found_from_seadex_id(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?perPage=500&expand=trs&filter=id%3D%27jashdsjhdsakjdhsakdhsadhaksjd%27&skipTotal=true",
        json={"page": 1, "perPage": 30, "totalItems": 0, "totalPages": 0, "items": []},
    )

    with pytest.raises(EntryNotFoundError):
        seadex_entry.from_id("jashdsjhdsakjdhsakdhsadhaksjd")


def test_entry_not_found_from_title(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://graphql.anilist.co",
        json={
            "errors": [{"message": "Not Found.", "status": 404, "locations": [{"line": 1, "column": 28}]}],
            "data": {"Media": None},
        },
    )

    with pytest.raises(EntryNotFoundError):
        seadex_entry.from_title("jashdsjhdsakjdhsakdhsadhaksjd")


def test_from_filter_invalid_type(seadex_entry: SeaDexEntry) -> None:
    with pytest.raises(TypeError, match="'filter' must be a string, not object."):
        next(seadex_entry.from_filter(object()))  # type: ignore[arg-type]


def test_from_infohash_type(seadex_entry: SeaDexEntry) -> None:
    with pytest.raises(TypeError, match="'infohash' must be a string, not object."):
        next(seadex_entry.from_infohash(object()))  # type: ignore[arg-type]


@pytest.mark.parametrize(
    "infohash",
    ["certainly not a infohash", "2120e6faea8860ffa07bf535ea89a150b4fda773a", "2120e6faea8860ffa07bf535ea89a150b4fda7"],
)
def test_from_infohash_format(seadex_entry: SeaDexEntry, infohash: str) -> None:
    with pytest.raises(ValueError, match="Invalid infohash format. Must be a 40-character hexadecimal string."):
        next(seadex_entry.from_infohash(infohash))



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from seadex import SeaDexBackup, SeaDexEntry

if TYPE_CHECKING:
    from collections.abc import Iterator

    from pytest_httpx import HTTPXMock


@pytest.fixture
def seadex_entry() -> Iterator[SeaDexEntry]:
    with SeaDexEntry() as seadex:
        yield seadex


@pytest.fixture
def seadex_backup(httpx_mock: HTTPXMock) -> Iterator[SeaDexBackup]:
    httpx_mock.add_response(url="https://releases.moe/api/admins/auth-with-password", json={"token": "secret"})
    httpx_mock.add_response(
        url="https://releases.moe/api/files/token", json={"token": "secret"}, is_reusable=True, is_optional=True
    )
    with SeaDexBackup("me@example.com", "example") as seadex:
        yield seadex



from __future__ import annotations

import re
from os.path import basename
from typing import TYPE_CHECKING, Any
from urllib.parse import urljoin

import httpx

from seadex._exceptions import EntryNotFoundError
from seadex._types import EntryRecord, StrPath
from seadex._utils import httpx_client

if TYPE_CHECKING:
    from collections.abc import Iterator

    from httpx import Client
    from typing_extensions import Self


class SeaDexEntry:
    def __init__(self, base_url: str = "https://releases.moe", *, client: Client | None = None) -> None:
        """
        Client to interact with SeaDex entries.

        Parameters
        ----------
        base_url : str, optional
            The base URL of SeaDex, used for constructing API queries.
        client : Client, optional
            An [`httpx.Client`][httpx.Client] instance used to make requests to SeaDex.

            [httpx.Client]: https://www.python-httpx.org/advanced/#client

        Examples
        --------
        ```py
        with SeaDexEntry() as entry:
            tamako = entry.from_title("tamako love story")
            for torrent in tamako.torrents:
                if torrent.is_best and torrent.tracker.is_public():
                    print(torrent.release_group)
                    #> LYS1TH3A
                    #> Okay-Subs
        ```

        """
        self._base_url = base_url
        self._endpoint = urljoin(self._base_url, "/api/collections/entries/records")
        self._client = httpx_client() if client is None else client
        # Internal cache for AniList ID and title lookup by search term.
        # Used to avoid repeated AniList API calls for the same title search.
        self._al_cache: dict[str, dict[str, Any]] = {}

    @property
    def base_url(self) -> str:
        """
        Base URL, used for constructing API queries.
        """
        return self._base_url

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: object) -> None:
        self.close()

    def __from_filter(self, filter: str | None, /, *, paginate: bool) -> Iterator[EntryRecord]:
        """Yield entries that match the provided filter."""
        params: dict[str, Any] = {}

        if filter is None:
            params.update({"perPage": 500, "expand": "trs"})
        else:
            params.update({"perPage": 500, "expand": "trs", "filter": filter})

        if paginate:
            first_page = self._client.get(self._endpoint, params=params).raise_for_status()
            data = first_page.json()
            total_pages = data["totalPages"]

            for item in data["items"]:  # Page 1
                yield EntryRecord.from_dict(item)

            for page in range(2, total_pages + 1):  # Page 2 to total_pages
                params.update({"page": page})
                response = self._client.get(self._endpoint, params=params).raise_for_status()
                for item in response.json()["items"]:
                    yield EntryRecord.from_dict(item)
        else:
            params.update({"skipTotal": True})
            response = self._client.get(self._endpoint, params=params).raise_for_status()
            for item in response.json()["items"]:
                yield EntryRecord.from_dict(item)

    def close(self) -> None:
        """
        Close the underlying HTTP client connection.
        """
        self._client.close()

    def from_filter(self, filter: str, /) -> Iterator[EntryRecord]:
        """
        Yield entries from SeaDex that match the given filter expression.

        Refer to the `filter` argument in the [PocketBase API documentation][]
        for details on constructing valid filter expressions.

        [PocketBase API documentation]: https://pocketbase.io/docs/api-records/#listsearch-records

        Parameters
        ----------
        filter : str
            The filter expression.

        Yields
        ------
        EntryRecord
            The retrieved entry.

        Raises
        ------
        TypeError
            If `filter` is not a string.

        """
        if not isinstance(filter, str):
            errmsg = f"'filter' must be a string, not {type(filter).__name__}."
            raise TypeError(errmsg)

        yield from self.__from_filter(filter, paginate=True)

    def from_id(self, id: int | str, /) -> EntryRecord:
        """
        Retrieve an entry by its ID.

        Parameters
        ----------
        id : int | str
            The ID of the entry. Can be an AniList ID (integer)
            or a SeaDex database ID (string).

        Returns
        -------
        EntryRecord
            The retrieved entry.

        Raises
        ------
        EntryNotFoundError
            If no entry is found for the provided ID.

        """
        filter = f"alID={id}" if isinstance(id, int) else f"id='{id}'"
        entries = self.__from_filter(filter, paginate=False)

        try:
            return next(entries)
        except StopIteration:
            errmsg = f"No seadex entry found for id: {id}"
            raise EntryNotFoundError(errmsg) from None

    def from_title(self, title: str, /) -> EntryRecord:
        """
        Retrieve an entry by its anime title.

        Parameters
        ----------
        title : str
            The title of the anime to search for.

        Returns
        -------
        EntryRecord
            The retrieved entry.

        Raises
        ------
        EntryNotFoundError
            If no entry is found for the provided title.

        """
        title = title.strip()
        try:
            try:
                # Attempt to retrieve AniList ID from cache
                anilist_id = self._al_cache[title]["id"]
            except KeyError:
                # If not in cache, query the AniList GraphQL API
                response = self._client.post(
                    "https://graphql.anilist.co",
                    json={
                        "query": (
                            "query ($search: String!) "
                            "{ Media(search: $search, type: ANIME) { id title { english romaji } } }"
                        ),
                        "variables": {"search": title},
                    },
                ).raise_for_status()

                # Cache the response
                self._al_cache[title] = media = response.json()["data"]["Media"]
                anilist_id = media["id"]

            entries = self.__from_filter(f"alID={anilist_id}", paginate=False)
            return next(entries)

        except (StopIteration, TypeError, httpx.HTTPStatusError):
            errmsg = f"No seadex entry found for title: {title}"
            raise EntryNotFoundError(errmsg) from None

    def from_filename(self, filename: StrPath, /) -> Iterator[EntryRecord]:
        """
        Yield entries that contain a torrent with the specified filename.

        Parameters
        ----------
        filename : StrPath
            The filename to search for.

        Yields
        ------
        EntryRecord
            The retrieved entry.

        """
        yield from self.__from_filter(f'trs.files?~\'"name":"{basename(filename)}"\'', paginate=False)

    def from_infohash(self, infohash: str, /) -> Iterator[EntryRecord]:
        """
        Yield entries that contain a torrent with the specified infohash.

        Parameters
        ----------
        infohash : str
            The infohash to search for.

        Yields
        ------
        EntryRecord
            The retrieved entry.

        Raises
        ------
        TypeError
            If `infohash` is not a string.
        ValueError
            If `infohash` is not a 40-character hexadecimal string.

        """
        if not isinstance(infohash, str):
            errmsg = f"'infohash' must be a string, not {type(infohash).__name__}."
            raise TypeError(errmsg)

        infohash = infohash.lower().strip()

        if not re.match(r"^[0-9a-f]{40}$", infohash):
            errmsg = "Invalid infohash format. Must be a 40-character hexadecimal string."
            raise ValueError(errmsg)

        yield from self.__from_filter(f"trs.infoHash?='{infohash}'", paginate=False)

    def iterator(self) -> Iterator[EntryRecord]:
        """
        Lazily iterate over all the entries in SeaDex.

        Yields
        ------
        EntryRecord
            The retrieved entry.

        """
        yield from self.__from_filter(None, paginate=True)



from __future__ import annotations

import base64
from datetime import datetime

from seadex import EntryRecord, File, TorrentRecord, Tracker


def test_torrent_record() -> None:
    """
    Tests that `_from_dict` correctly parses a dictionary into a TorrentRecord object.
    """
    sample_data = {
        "collectionId": "oiwizhmushn5qqh",
        "collectionName": "torrents",
        "created": "2024-01-30 19:28:09.110Z",
        "dualAudio": True,
        "files": [
            {"length": 4636316199, "name": "Tamako.Love.Story.2014.1080p.BluRay.Opus2.0.H.265-LYS1TH3A.mkv"},
        ],
        "id": "pcpina3ekbqk7a5",
        "infoHash": "23f77120cfdf9df8b42a10216aa33e281c58b456",
        "isBest": True,
        "releaseGroup": "LYS1TH3A",
        "tracker": "Nyaa",
        "updated": "2024-01-30 19:28:09.110Z",
        "url": "https://nyaa.si/view/1693872",
    }
    record = TorrentRecord.from_dict(sample_data)

    assert TorrentRecord.from_json(record.to_json()) == TorrentRecord.from_dict(record.to_dict())
    assert record.collection_id == "oiwizhmushn5qqh"
    assert record.collection_name == "torrents"
    assert isinstance(record.created_at, datetime)
    assert record.is_dual_audio
    assert len(record.files) == 1
    assert record.files[0].name == "Tamako.Love.Story.2014.1080p.BluRay.Opus2.0.H.265-LYS1TH3A.mkv"
    assert record.files[0].size == 4636316199
    assert File.from_json(record.files[0].to_json()) == File.from_dict(record.files[0].to_dict())
    assert record.size == 4636316199
    assert record.id == "pcpina3ekbqk7a5"
    assert record.infohash == "23f77120cfdf9df8b42a10216aa33e281c58b456"
    assert record.is_best
    assert record.release_group == "LYS1TH3A"
    assert record.tracker is Tracker.NYAA
    assert isinstance(record.updated_at, datetime)
    assert record.url == "https://nyaa.si/view/1693872"


def test_entry_record() -> None:
    sample_data = {
        "alID": 20519,
        "collectionId": "3l2x9nxip35gqb5",
        "collectionName": "entries",
        "comparison": "https://slow.pics/c/rc6qrB1F",
        "created": "2024-01-30 19:28:10.337Z",
        "expand": {
            "trs": [
                {
                    "collectionId": "oiwizhmushn5qqh",
                    "collectionName": "torrents",
                    "created": "2024-01-30 19:28:09.110Z",
                    "dualAudio": True,
                    "files": [
                        {"length": 4636316199, "name": "Tamako.Love.Story.2014.1080p.BluRay.Opus2.0.H.265-LYS1TH3A.mkv"}
                    ],
                    "id": "pcpina3ekbqk7a5",
                    "infoHash": "23f77120cfdf9df8b42a10216aa33e281c58b456",
                    "isBest": True,
                    "releaseGroup": "LYS1TH3A",
                    "tracker": "Nyaa",
                    "updated": "2024-01-30 19:28:09.110Z",
                    "url": "https://nyaa.si/view/1693872",
                },
                {
                    "collectionId": "oiwizhmushn5qqh",
                    "collectionName": "torrents",
                    "created": "2024-01-30 19:28:09.461Z",
                    "dualAudio": True,
                    "files": [
                        {"length": 4636316199, "name": "Tamako.Love.Story.2014.1080p.BluRay.Opus2.0.H.265-LYS1TH3A.mkv"}
                    ],
                    "id": "tvh4fn4m2qi19n5",
                    "infoHash": "<redacted>",
                    "isBest": True,
                    "releaseGroup": "LYS1TH3A",
                    "tracker": "AB",
                    "updated": "2024-01-30 19:28:09.461Z",
                    "url": "/torrents.php?id=20684&torrentid=1053072",
                },
            ]
        },
        "id": "c344w8ld7q1yppz",
        "incomplete": False,
        "notes": "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub",
        "theoreticalBest": "",
        "trs": ["pcpina3ekbqk7a5", "tvh4fn4m2qi19n5", "qhcmujh4dsw55j2", "enytf1g1cxf0k47"],
        "updated": "2024-01-30 19:28:10.337Z",
    }

    record = EntryRecord.from_dict(sample_data)

    assert EntryRecord.from_json(record.to_json()) == EntryRecord.from_dict(record.to_dict())
    assert record.anilist_id == 20519
    assert record.collection_id == "3l2x9nxip35gqb5"
    assert record.collection_name == "entries"
    assert record.comparisons == ("https://slow.pics/c/rc6qrB1F",)  # Tuple with one element
    assert isinstance(record.created_at, datetime)
    assert record.id == "c344w8ld7q1yppz"
    assert not record.is_incomplete
    assert record.notes == (
        "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub"
    )
    assert record.theoretical_best is None
    assert record.torrents[0].url == "https://nyaa.si/view/1693872"
    assert (
        record.torrents[1].url
        == base64.b64decode(
            b"aHR0cHM6Ly9hbmltZWJ5dGVzLnR2L3RvcnJlbnRzLnBocD9pZD0yMDY4NCZ0b3JyZW50aWQ9MTA1MzA3Mg=="
        ).decode()
    )
    assert record.torrents[0].infohash is not None
    assert record.torrents[1].infohash is None
    assert isinstance(record.updated_at, datetime)
    assert record.url == "https://releases.moe/20519/"
    assert record.size == sum(tr.size for tr in record.torrents)



from __future__ import annotations

import base64

import pytest

from seadex import Tracker


@pytest.mark.parametrize(
    "tracker, value",
    [
        (Tracker.NYAA, "Nyaa"),
        (Tracker.ANIMETOSHO, "AnimeTosho"),
        (Tracker.ANIDEX, "AniDex"),
        (Tracker.RUTRACKER, "RuTracker"),
        (Tracker.ANIMEBYTES, "AB"),
        (Tracker.BEYONDHD, "BeyondHD"),
        (Tracker.PASSTHEPOPCORN, "PassThePopcorn"),
        (Tracker.BROADCASTTHENET, "BroadcastTheNet"),
        (Tracker.HDBITS, "HDBits"),
        (Tracker.BLUTOPIA, "Blutopia"),
        (Tracker.AITHER, "Aither"),
        (Tracker.OTHER, "Other"),
        (Tracker.OTHER_PRIVATE, "OtherPrivate"),
    ],
)
def test_tracker_values(tracker: Tracker, value: str) -> None:
    assert tracker == value


@pytest.mark.parametrize(
    "tracker, is_private, is_public",
    [
        ("Nyaa", False, True),
        ("AnimeTosho", False, True),
        ("AniDex", False, True),
        ("RuTracker", False, True),
        ("AB", True, False),
        ("BeyondHD", True, False),
        ("PassThePopcorn", True, False),
        ("BroadcastTheNet", True, False),
        ("HDBits", True, False),
        ("Blutopia", True, False),
        ("Aither", True, False),
        ("Other", False, True),
        ("OtherPrivate", True, False),
    ],
)
def test_tracker_is_private(tracker: str, is_private: bool, is_public: bool) -> None:
    assert Tracker(tracker).is_private() == is_private
    assert Tracker(tracker).is_public() == is_public


@pytest.mark.parametrize(
    "tracker, url",
    [
        ("Nyaa", b"aHR0cHM6Ly9ueWFhLnNp"),
        ("AnimeTosho", b"aHR0cHM6Ly9hbmltZXRvc2hvLm9yZw=="),
        ("AniDex", b"aHR0cHM6Ly9hbmlkZXguaW5mbw=="),
        ("RuTracker", b"aHR0cHM6Ly9ydXRyYWNrZXIub3Jn"),
        ("AB", b"aHR0cHM6Ly9hbmltZWJ5dGVzLnR2"),
        ("BeyondHD", b"aHR0cHM6Ly9iZXlvbmQtaGQubWU="),
        ("PassThePopcorn", b"aHR0cHM6Ly9wYXNzdGhlcG9wY29ybi5tZQ=="),
        ("BroadcastTheNet", b"aHR0cHM6Ly9icm9hZGNhc3RoZS5uZXQ="),
        ("HDBits", b"aHR0cHM6Ly9oZGJpdHMub3Jn"),
        ("Blutopia", b"aHR0cHM6Ly9ibHV0b3BpYS5jYw=="),
        ("Aither", b"aHR0cHM6Ly9haXRoZXIuY2M="),
        ("Other", b""),
        ("OtherPrivate", b""),
    ],
)
def test_tracker_domain(tracker: str, url: bytes) -> None:
    assert Tracker(tracker).url == base64.b64decode(url).decode()


def test_bad_value() -> None:
    with pytest.raises(ValueError):
        Tracker("kasdjsahdjshdahdakjds")

    with pytest.raises(ValueError):
        Tracker(121212)  # type: ignore[arg-type]


def test_case_insensitive_lookup() -> None:
    assert Tracker("nyAA") is Tracker.NYAA
    assert Tracker("ANIMETOSHO") is Tracker.ANIMETOSHO
    assert Tracker("Ab") is Tracker.ANIMEBYTES



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, TypeAlias

from cyclopts import App, Parameter
from cyclopts.types import ResolvedExistingDirectory

if TYPE_CHECKING:
    from seadex._backup import SeaDexBackup

backup_app = App(
    "backup",
    help="Perform backup operations.",
    help_format="plaintext",
)

EmailType: TypeAlias = Annotated[str, Parameter(env_var="SEADEX_EMAIL")]
PasswordType: TypeAlias = Annotated[str, Parameter(env_var="SEADEX_PASSWORD")]


def _login(email: str, password: str) -> SeaDexBackup:
    """
    Log in to SeaDex using email and password.

    Parameters
    ----------
    email : str
        The administrator email used for authentication.
    password : str
        The administrator password used for authentication.

    """
    from rich.console import Console

    from seadex._backup import SeaDexBackup

    with Console().status("Logging in", spinner="earth"):
        return SeaDexBackup(email, password)


@backup_app.command(name="list")
def _list(*, email: EmailType, password: PasswordType, json: bool = False) -> None:
    """
    List all available backups.

    Parameters
    ----------
    email : str
        The administrator email used for authentication.
    password : str
        The administrator password used for authentication.
    json : bool, optional
        If True, the output will be a JSON string.

    """
    from humanize import naturalsize
    from rich import box, print, print_json
    from rich.table import Table

    client = _login(email, password)
    backups = client.get_backups()

    if json:
        import msgspec

        jsonified = msgspec.json.encode(backups).decode()
        print_json(jsonified)
        return

    table = Table("Name", "Size", "Date Modified", box=box.ROUNDED)
    for backup in backups:
        table.add_row(backup.name, naturalsize(backup.size), backup.modified_time.isoformat())
    print(table)


@backup_app.command
def create(name: str | None = None, /, *, email: EmailType, password: PasswordType) -> None:
    """
    Create a new backup.

    Parameters
    ----------
    name : str, optional
        The name of the backup. If not provided, a default name is generated using the
        template `%Y%m%d%H%M%S-seadex-backup.zip`, which supports the full `datetime.strftime`
        formatting.
    email : EmailType
        The administrator email used for authentication.
    password : PasswordType
        The administrator password used for authentication.

    """
    from rich.console import Console

    console = Console()
    client = _login(email, password)
    filename = "%Y%m%d%H%M%S-seadex-backup.zip" if name is None else name

    with console.status("Creating a backup"):
        backup = client.create(filename)
    console.print(f":package: Created {backup}", emoji=True, highlight=False)


@backup_app.command
def download(
    name: str | None = None,
    /,
    *,
    email: EmailType,
    password: PasswordType,
    destination: ResolvedExistingDirectory | None = None,
    existing: bool = True,
) -> None:
    """
    Download a backup.

    Parameters
    ----------
    name : str, optional
        The name of the backup to download. If not provided, the latest backup is downloaded.
    email : EmailType
        The administrator email used for authentication.
    password : PasswordType
        The administrator password used for authentication.
    destination : ResolvedExistingDirectory | None, optional
        The destination directory for the backup. Defaults to the current working directory.
    existing : bool, optional
        If `True`, download an existing backup. If `False`, create a temporary backup on the remote system,
        download it, and then delete it from the remote.

    """
    from rich.console import Console

    console = Console()
    client = _login(email, password)

    if not existing:
        if name is None:
            console.print("[red]error:[/] The `--name` option is required when using `--no-existing`.")
            return

        with console.status("Creating a temporary backup on remote"):
            backup = client.create(name)
        console.print(f":white_check_mark: Created [cyan]{backup}[/cyan] on remote", emoji=True)

        with console.status(f"Downloading [cyan]{backup}[/cyan]"):
            dest = client.download(backup, destination=destination)
        console.print(f":package: Saved to [cyan]{dest}[/cyan]", emoji=True)

        with console.status(f"Deleting [cyan]{backup}[/cyan] from remote"):
            client.delete(backup)
        console.print(f":litter_in_bin_sign: Deleted [cyan]{backup}[/cyan] from remote", emoji=True)

        return

    backup = name or client.get_latest_backup()  # type: ignore[assignment]
    with console.status(f"Downloading [cyan]{backup}[/cyan]"):
        dest = client.download(backup, destination=destination)
    console.print(f":package: Saved to [cyan]{dest}[/cyan]", emoji=True, highlight=False)


@backup_app.command
def delete(name: str, /, *, email: EmailType, password: PasswordType) -> None:
    """
    Delete a backup by name.

    Parameters
    ----------
    name : str
        The name of the backup to delete.
    email : EmailType
        The administrator email used for authentication.
    password : PasswordType
        The administrator password used for authentication.

    """
    from rich.console import Console

    console = Console()

    client = _login(email, password)

    with console.status(f"Deleting {name}"):
        client.delete(name)

    console.print(f":litter_in_bin_sign: Deleted {name}", emoji=True, highlight=False)



from __future__ import annotations


class SeaDexError(Exception):
    """Base Exception for all SeaDex related errors."""

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class EntryNotFoundError(SeaDexError):
    """The requested Entry was not found in SeaDex."""


class BadBackupFileError(SeaDexError):
    """The backup file is broken."""



from __future__ import annotations

import shutil
from pathlib import Path

import pytest
from torf import Torrent

from seadex import SeaDexTorrent


def test_sanitize_file_exists_error() -> None:
    file = "tests/__torrents__/private-ubuntu-24.04.1-desktop-amd64.iso.torrent"
    with pytest.raises(FileExistsError):
        SeaDexTorrent(file).sanitize()


def test_sanitize_private_torrent(tmp_path: Path) -> None:
    file = "tests/__torrents__/private-ubuntu-24.04.1-desktop-amd64.iso.torrent"
    original = Torrent.read(file)
    sanitized = Torrent.read(
        SeaDexTorrent(file).sanitize(destination=tmp_path / "private-ubuntu-24.04.1-desktop-amd64.iso.torrent")
    )

    assert original.trackers is not None
    assert sanitized.trackers == []

    assert original.webseeds is not None
    assert sanitized.webseeds == []

    assert original.httpseeds is not None
    assert sanitized.httpseeds == []

    assert original.private is True
    assert sanitized.private is None

    assert original.comment is not None
    assert sanitized.comment is None

    assert original.creation_date is not None
    assert sanitized.creation_date is None

    assert original.created_by is not None
    assert sanitized.created_by is None

    assert original.source is not None
    assert sanitized.source is None

    assert original.infohash != sanitized.infohash
    assert original.infohash_base32 != sanitized.infohash_base32


def test_sanitize_private_torrent_with_overwrite(tmp_path: Path) -> None:
    src = "tests/__torrents__/private-ubuntu-24.04.1-desktop-amd64.iso.torrent"
    testfile = Path(shutil.copy(src, tmp_path))

    assert testfile.is_file()

    original = Torrent.read(testfile)
    sanitized = Torrent.read(SeaDexTorrent(testfile).sanitize(overwrite=True))

    assert original.trackers is not None
    assert sanitized.trackers == []

    assert original.webseeds is not None
    assert sanitized.webseeds == []

    assert original.httpseeds is not None
    assert sanitized.httpseeds == []

    assert original.private is True
    assert sanitized.private is None

    assert original.comment is not None
    assert sanitized.comment is None

    assert original.creation_date is not None
    assert sanitized.creation_date is None

    assert original.created_by is not None
    assert sanitized.created_by is None

    assert original.source is not None
    assert sanitized.source is None

    assert original.infohash != sanitized.infohash
    assert original.infohash_base32 != sanitized.infohash_base32


def test_sanitize_public_torrent(tmp_path: Path) -> None:
    file = "tests/__torrents__/public-ubuntu-24.04.1-desktop-amd64.iso.torrent"
    original = Torrent.read(file)
    sanitized = Torrent.read(
        SeaDexTorrent(file).sanitize(
            destination=tmp_path / "public-ubuntu-24.04.1-desktop-amd64.iso.torrent", overwrite=True
        )
    )

    assert original.trackers == sanitized.trackers
    assert original.webseeds == sanitized.webseeds
    assert original.httpseeds == sanitized.httpseeds
    assert original.private == sanitized.private
    assert original.comment == sanitized.comment
    assert original.creation_date == sanitized.creation_date
    assert original.created_by == sanitized.created_by
    assert original.source == sanitized.source
    assert original.infohash == sanitized.infohash
    assert original.infohash_base32 == sanitized.infohash_base32


def test_filelist_public() -> None:
    file = "tests/__torrents__/public-ubuntu-24.04.1-desktop-amd64.iso.torrent"

    original = Torrent.read(file)
    seadex = SeaDexTorrent(file)

    original_files = [{"filename": Path(file).as_posix(), "size": int(file.size)} for file in original.files]
    seadex_filelist = [{"filename": file.name, "size": file.size} for file in seadex.filelist]
    seadex_filelist_str = [{"filename": str(file), "size": file.size} for file in seadex.filelist]

    assert original_files == seadex_filelist == seadex_filelist_str


def test_filelist_private() -> None:
    file = "tests/__torrents__/private-ubuntu-24.04.1-desktop-amd64.iso.torrent"

    original = Torrent.read(file)
    seadex = SeaDexTorrent(file)

    original_files = [{"filename": Path(file).as_posix(), "size": int(file.size)} for file in original.files]
    seadex_filelist = [{"filename": file.name, "size": file.size} for file in seadex.filelist]
    seadex_filelist_str = [{"filename": str(file), "size": file.size} for file in seadex.filelist]

    assert original_files == seadex_filelist == seadex_filelist_str



from __future__ import annotations

from cyclopts import App

entry_app = App(
    "entry",
    help="Get a SeaDex entry.",
    help_format="plaintext",
)


@entry_app.default
def get_entry(title: str, /, *, json: bool = False) -> None:
    """
    Get the seadex entry for the given title.

    Parameters
    ----------
    title : str
        Title of the anime.
    json : bool, optional
        If True, the output will be a JSON string.

    """
    from rich import box, print_json
    from rich.console import Console, Group
    from rich.table import Table
    from rich.theme import Theme

    from seadex._entry import SeaDexEntry
    from seadex._exceptions import EntryNotFoundError

    console = Console(theme=Theme({"repr.number": ""}))

    with SeaDexEntry() as seadex_entry:
        try:
            entry = seadex_entry.from_title(title)
            # Grab the anilist title from internal cache
            # This saves us from hitting AniList again
            anilist_title = (
                seadex_entry._al_cache[title]["title"]["english"] or seadex_entry._al_cache[title]["title"]["romaji"]
            )
        except EntryNotFoundError:
            console.print(f"[red]error:[/] no seadex entry found for {title!r}")
            return

        if json:
            print_json(entry.to_json())
            return

        body = f"Title: {anilist_title}\n"
        body += f"URL: {entry.url}\n"
        body += f"AniList: https://anilist.co/anime/{entry.anilist_id}\n"
        body += f"Incomplete: {'Yes' if entry.is_incomplete else 'No'}\n"
        body += f"Updated At: {entry.updated_at.strftime('%b %d, %Y')}"
        if entry.theoretical_best is not None:
            body += f"\nTheoretical Best: {entry.theoretical_best}"

        table = Table(box=box.ROUNDED)
        table.add_column("Group")
        table.add_column("Best")
        table.add_column("Dual")
        table.add_column("URL")

        for torrent in sorted(
            entry.torrents, key=lambda t: (not (t.is_best and t.is_dual_audio), not t.is_best, not t.is_dual_audio)
        ):
            table.add_row(
                torrent.release_group,
                ":white_check_mark:" if torrent.is_best else ":cross_mark:",
                ":white_check_mark:" if torrent.is_dual_audio else ":cross_mark:",
                torrent.url,
            )

        lower_body = f"Notes:\n{entry.notes}\n"
        lower_body += "\nComparisons:\n"
        for comparison in entry.comparisons:
            lower_body += f"- {comparison}\n"

        console.print(Group(body, table, lower_body))



from __future__ import annotations

from seadex._backup import BackupFile, SeaDexBackup
from seadex._entry import SeaDexEntry
from seadex._enums import Tracker
from seadex._exceptions import BadBackupFileError, EntryNotFoundError, SeaDexError
from seadex._torrent import SeaDexTorrent
from seadex._types import EntryRecord, File, TorrentRecord
from seadex._version import __version__

__all__ = (
    "BackupFile",
    "BadBackupFileError",
    "EntryNotFoundError",
    "EntryRecord",
    "File",
    "SeaDexBackup",
    "SeaDexEntry",
    "SeaDexError",
    "SeaDexTorrent",
    "TorrentRecord",
    "Tracker",
    "__version__",
)



from __future__ import annotations

import base64
import copy
from datetime import datetime, timezone
from typing import TYPE_CHECKING

from seadex import SeaDexEntry, Tracker

if TYPE_CHECKING:
    from pytest_httpx import HTTPXMock

SAMPLE_JSON_REPLY = {
    "page": 1,
    "perPage": 500,
    "totalItems": 1,
    "totalPages": 1,
    "items": [
        {
            "alID": 20519,
            "collectionId": "3l2x9nxip35gqb5",
            "collectionName": "entries",
            "comparison": "https://slow.pics/c/rc6qrB1F",
            "created": "2024-01-30 19:28:10.337Z",
            "expand": {
                "trs": [
                    {
                        "collectionId": "oiwizhmushn5qqh",
                        "collectionName": "torrents",
                        "created": "2024-01-30 19:28:09.110Z",
                        "dualAudio": True,
                        "files": [
                            {
                                "length": 4636316199,
                                "name": "Tamako.Love.Story.2014.1080p.BluRay.Opus2.0.H.265-LYS1TH3A.mkv",
                            }
                        ],
                        "id": "pcpina3ekbqk7a5",
                        "infoHash": "23f77120cfdf9df8b42a10216aa33e281c58b456",
                        "isBest": True,
                        "releaseGroup": "LYS1TH3A",
                        "tracker": "Nyaa",
                        "updated": "2024-01-30 19:28:09.110Z",
                        "url": "https://nyaa.si/view/1693872",
                    },
                    {
                        "collectionId": "oiwizhmushn5qqh",
                        "collectionName": "torrents",
                        "created": "2024-01-30 19:28:09.461Z",
                        "dualAudio": True,
                        "files": [
                            {
                                "length": 4636316199,
                                "name": "Tamako.Love.Story.2014.1080p.BluRay.Opus2.0.H.265-LYS1TH3A.mkv",
                            }
                        ],
                        "id": "tvh4fn4m2qi19n5",
                        "infoHash": "<redacted>",
                        "isBest": True,
                        "releaseGroup": "LYS1TH3A",
                        "tracker": "AB",
                        "updated": "2024-01-30 19:28:09.461Z",
                        "url": "/torrents.php?id=20684&torrentid=1053072",
                    },
                    {
                        "collectionId": "oiwizhmushn5qqh",
                        "collectionName": "torrents",
                        "created": "2024-01-30 19:28:09.803Z",
                        "dualAudio": False,
                        "files": [{"length": 4555215904, "name": "[Okay-Subs] Tamako Love Story [45C10FA7].mkv"}],
                        "id": "qhcmujh4dsw55j2",
                        "infoHash": "cfb670c2261701b060b708a16743cf6658c47b62",
                        "isBest": True,
                        "releaseGroup": "Okay-Subs",
                        "tracker": "Nyaa",
                        "updated": "2024-01-30 19:28:09.803Z",
                        "url": "https://nyaa.si/view/1656471",
                    },
                    {
                        "collectionId": "oiwizhmushn5qqh",
                        "collectionName": "torrents",
                        "created": "2024-01-30 19:28:10.159Z",
                        "dualAudio": False,
                        "files": [{"length": 4555215904, "name": "[Okay-Subs] Tamako Love Story [45C10FA7].mkv"}],
                        "id": "enytf1g1cxf0k47",
                        "infoHash": "<redacted>",
                        "isBest": True,
                        "releaseGroup": "Okay-Subs",
                        "tracker": "AB",
                        "updated": "2024-01-30 19:28:10.159Z",
                        "url": "/torrents.php?id=20684&torrentid=1031817",
                    },
                ]
            },
            "id": "c344w8ld7q1yppz",
            "incomplete": False,
            "notes": "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub",
            "theoreticalBest": "",
            "trs": ["pcpina3ekbqk7a5", "tvh4fn4m2qi19n5", "qhcmujh4dsw55j2", "enytf1g1cxf0k47"],
            "updated": "2024-01-30 19:28:10.337Z",
        }
    ],
}


def test_properties(seadex_entry: SeaDexEntry) -> None:
    assert seadex_entry.base_url == "https://releases.moe"


def test_from_anilist_id(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?perPage=500&expand=trs&filter=alID%3D20519&skipTotal=true",
        json=SAMPLE_JSON_REPLY,
    )
    entry = seadex_entry.from_id(20519)
    assert entry.anilist_id == 20519
    assert entry.collection_id == "3l2x9nxip35gqb5"
    assert entry.collection_name == "entries"
    assert entry.comparisons == ("https://slow.pics/c/rc6qrB1F",)
    assert entry.created_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)
    assert entry.id == "c344w8ld7q1yppz"
    assert not entry.is_incomplete
    assert (
        entry.notes == "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub"
    )
    assert entry.theoretical_best is None
    assert entry.torrents[0].url == "https://nyaa.si/view/1693872"
    assert (
        entry.torrents[1].url
        == base64.b64decode(
            b"aHR0cHM6Ly9hbmltZWJ5dGVzLnR2L3RvcnJlbnRzLnBocD9pZD0yMDY4NCZ0b3JyZW50aWQ9MTA1MzA3Mg=="
        ).decode()
    )
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.torrents[0].tracker is Tracker.NYAA
    assert entry.torrents[1].tracker is Tracker.ANIMEBYTES
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert isinstance(entry.updated_at, datetime)


def test_from_seadex_id(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?perPage=500&expand=trs&filter=id%3D%27c344w8ld7q1yppz%27&skipTotal=true",
        json=SAMPLE_JSON_REPLY,
    )
    entry = seadex_entry.from_id("c344w8ld7q1yppz")
    assert entry.anilist_id == 20519
    assert entry.collection_id == "3l2x9nxip35gqb5"
    assert entry.collection_name == "entries"
    assert entry.comparisons == ("https://slow.pics/c/rc6qrB1F",)
    assert entry.created_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)
    assert entry.id == "c344w8ld7q1yppz"
    assert not entry.is_incomplete
    assert (
        entry.notes == "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub"
    )
    assert entry.theoretical_best is None
    assert entry.torrents[0].url == "https://nyaa.si/view/1693872"
    assert (
        entry.torrents[1].url
        == base64.b64decode(
            b"aHR0cHM6Ly9hbmltZWJ5dGVzLnR2L3RvcnJlbnRzLnBocD9pZD0yMDY4NCZ0b3JyZW50aWQ9MTA1MzA3Mg=="
        ).decode()
    )
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.torrents[0].tracker is Tracker.NYAA
    assert entry.torrents[1].tracker is Tracker.ANIMEBYTES
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.updated_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)


def test_from_title(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://graphql.anilist.co",
        json={
            "data": {"Media": {"id": 20519, "title": {"english": "Tamako -love story-", "romaji": "Tamako Love Story"}}}
        },
    )

    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?perPage=500&expand=trs&filter=alID%3D20519&skipTotal=true",
        json=SAMPLE_JSON_REPLY,
    )

    entry = seadex_entry.from_title("tamako love story")
    assert entry.anilist_id == 20519
    assert entry.collection_id == "3l2x9nxip35gqb5"
    assert entry.collection_name == "entries"
    assert entry.comparisons == ("https://slow.pics/c/rc6qrB1F",)
    assert entry.created_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)
    assert entry.id == "c344w8ld7q1yppz"
    assert not entry.is_incomplete
    assert (
        entry.notes == "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub"
    )
    assert entry.theoretical_best is None
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.torrents[0].tracker is Tracker.NYAA
    assert entry.torrents[1].tracker is Tracker.ANIMEBYTES
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.updated_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)


def test_from_filename(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?filter=trs.files%3F~%27%22name%22%3A%22Tamako.Love.Story.2014.1080p.BluRay.Opus2.0.H.265-LYS1TH3A.mkv%22%27&perPage=500&expand=trs&&skipTotal=true",
        json=SAMPLE_JSON_REPLY,
    )

    entries = seadex_entry.from_filename("Tamako.Love.Story.2014.1080p.BluRay.Opus2.0.H.265-LYS1TH3A.mkv")
    entry = tuple(entries)[0]  # noqa: RUF015
    assert entry.anilist_id == 20519
    assert entry.collection_id == "3l2x9nxip35gqb5"
    assert entry.collection_name == "entries"
    assert entry.comparisons == ("https://slow.pics/c/rc6qrB1F",)
    assert entry.created_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)
    assert entry.id == "c344w8ld7q1yppz"
    assert not entry.is_incomplete
    assert (
        entry.notes == "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub"
    )
    assert entry.theoretical_best is None
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.torrents[0].tracker is Tracker.NYAA
    assert entry.torrents[1].tracker is Tracker.ANIMEBYTES
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.updated_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)


def test_from_infohash(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?perPage=500&expand=trs&filter=trs.infoHash%3F%3D%2723f77120cfdf9df8b42a10216aa33e281c58b456%27&skipTotal=true",
        json=SAMPLE_JSON_REPLY,
    )

    entries = tuple(seadex_entry.from_infohash("23f77120cfdf9df8b42a10216aa33e281c58b456"))
    assert len(entries) == 1
    entry = tuple(entries)[0]  # noqa: RUF015
    assert entry.anilist_id == 20519
    assert entry.collection_id == "3l2x9nxip35gqb5"
    assert entry.collection_name == "entries"
    assert entry.comparisons == ("https://slow.pics/c/rc6qrB1F",)
    assert entry.created_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)
    assert entry.id == "c344w8ld7q1yppz"
    assert not entry.is_incomplete
    assert (
        entry.notes == "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub"
    )
    assert entry.theoretical_best is None
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.torrents[0].tracker is Tracker.NYAA
    assert entry.torrents[1].tracker is Tracker.ANIMEBYTES
    assert entry.torrents[0].infohash is not None
    assert entry.torrents[1].infohash is None
    assert entry.updated_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)


def test_iterator(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    # Mimic multi page response
    SAMPLE_JSON_REPLY_PAGE_1 = copy.deepcopy(SAMPLE_JSON_REPLY)  # noqa: N806
    SAMPLE_JSON_REPLY_PAGE_2 = copy.deepcopy(SAMPLE_JSON_REPLY)  # noqa: N806

    # First page should report 2 total pages
    SAMPLE_JSON_REPLY_PAGE_1["totalPages"] = 2

    # Second page should report itself as page 2
    SAMPLE_JSON_REPLY_PAGE_2["page"] = 2
    SAMPLE_JSON_REPLY_PAGE_2["totalPages"] = 2

    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?perPage=500&expand=trs",
        json=SAMPLE_JSON_REPLY_PAGE_1,
    )

    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?perPage=500&expand=trs&page=2",
        json=SAMPLE_JSON_REPLY_PAGE_2,
    )

    for entry in seadex_entry.iterator():
        assert entry.anilist_id == 20519
        assert entry.collection_id == "3l2x9nxip35gqb5"
        assert entry.collection_name == "entries"
        assert entry.comparisons == ("https://slow.pics/c/rc6qrB1F",)
        assert entry.created_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)
        assert entry.id == "c344w8ld7q1yppz"
        assert not entry.is_incomplete
        assert (
            entry.notes
            == "Okay-Subs is JPN BD Encode+Commie with additional honorifics track\nLYS1TH3A is Okay-Subs+Dub"
        )
        assert entry.theoretical_best is None
        assert entry.torrents[0].url == "https://nyaa.si/view/1693872"
        assert (
            entry.torrents[1].url
            == base64.b64decode(
                b"aHR0cHM6Ly9hbmltZWJ5dGVzLnR2L3RvcnJlbnRzLnBocD9pZD0yMDY4NCZ0b3JyZW50aWQ9MTA1MzA3Mg=="
            ).decode()
        )
        assert entry.torrents[0].infohash is not None
        assert entry.torrents[1].infohash is None
        assert entry.torrents[0].tracker is Tracker.NYAA
        assert entry.torrents[1].tracker is Tracker.ANIMEBYTES
        assert entry.torrents[0].infohash is not None
        assert entry.torrents[1].infohash is None
        assert entry.updated_at == datetime(2024, 1, 30, 19, 28, 10, 337000, tzinfo=timezone.utc)


def test_from_filter(seadex_entry: SeaDexEntry, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?perPage=500&expand=trs&filter=alID=20519",
        json=SAMPLE_JSON_REPLY,
    )

    httpx_mock.add_response(
        url="https://releases.moe/api/collections/entries/records?perPage=500&expand=trs&filter=alID%3D20519&skipTotal=true",
        json=SAMPLE_JSON_REPLY,
    )

    assert next(seadex_entry.from_filter(f"alID={20519}")) == seadex_entry.from_id(20519)



from __future__ import annotations

from cyclopts import App
from cyclopts.types import ResolvedExistingPath, ResolvedPath

torrent_app = App(
    "torrent",
    help="Perform torrent operations.",
    help_format="plaintext",
)


@torrent_app.command
def sanitize(
    src: ResolvedExistingPath,
    dst: ResolvedPath | None = None,
    /,
    *,
    overwrite: bool = False,
) -> None:
    """
    Sanitize torrent files by removing sensitive data.

    Parameters
    ----------
    src : ResolvedExistingPath
        Path to the source torrent file to sanitize.
    dst : ResolvedPath or None, optional
        Path where the sanitized file will be stored.
    overwrite: bool, optional
        If True, overwrites the destination file if it exists.

    """
    from rich.console import Console

    from seadex._torrent import SeaDexTorrent

    console = Console()

    if src.is_file() and src.suffix.lower() == ".torrent":
        try:
            destination = SeaDexTorrent(src).sanitize(destination=dst, overwrite=overwrite)
        except FileExistsError:
            console.print(
                "[red]error:[/] destination file already exists and overwrite is false. "
                "Use the `--overwrite` flag to replace it."
            )
            return
        console.print(f":white_check_mark: Saved sanitized torrent to [cyan]{destination}[/cyan]", emoji=True)
        return
    console.print("[red]error:[/] src must be an existing torrent file.")


@torrent_app.command
def filelist(src: ResolvedExistingPath, /, *, json: bool = False) -> None:
    """
    Output the list of files in a torrent.

    Parameters
    ----------
    src : ResolvedExistingPath
        Path to the torrent file.
    json : bool, optional
        If True, the output will be a SeaDex compatible JSON string.

    """

    from rich import box, print, print_json

    from seadex._torrent import SeaDexTorrent

    filelist = SeaDexTorrent(src).filelist

    if json:
        import msgspec

        jsonified = msgspec.json.encode(filelist).decode()
        print_json(jsonified)
        return

    import os

    from humanize import naturalsize
    from rich.table import Table

    table = Table("Filename", "Size", box=box.ROUNDED)
    parent = os.path.commonpath(file.name for file in filelist)

    for file in filelist:
        table.add_row(os.path.relpath(file.name, start=parent), naturalsize(file.size))

    print(table)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import httpx

from seadex._version import __version__

if TYPE_CHECKING:
    from seadex._types import StrPath


def realpath(path: StrPath) -> Path:
    """
    Resolve Path or Path-like strings and return a Path object.
    """
    return Path(path).expanduser().resolve()


def httpx_client(*, timeout: httpx.Timeout | None = None) -> httpx.Client:
    """
    Return an instance of an httpx.Client.
    """
    headers = {"User-Agent": f"seadex/{__version__} (https://pypi.org/project/seadex/)"}

    if timeout is not None:
        return httpx.Client(headers=headers, timeout=timeout)
    return httpx.Client(headers=headers)



from __future__ import annotations

import base64
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from enum import Enum

    class StrEnum(str, Enum):
        pass


class CaseInsensitiveStrEnum(StrEnum):
    """StrEnum with case-insensitive lookup."""

    @classmethod
    def _missing_(cls, value: object) -> Self:
        # https://docs.python.org/3/library/enum.html#enum.Enum._missing_
        msg = f"'{value}' is not a valid {cls.__name__}"

        if isinstance(value, str):
            for member in cls:
                if member.value.lower() == value.lower():
                    return member
            raise ValueError(msg)
        raise ValueError(msg)


class Tracker(CaseInsensitiveStrEnum):
    """Enum of public and private trackers."""

    # Public Trackers
    NYAA = "Nyaa"
    ANIMETOSHO = "AnimeTosho"
    ANIDEX = "AniDex"
    RUTRACKER = "RuTracker"
    # Private Trackers
    ANIMEBYTES = "AB"
    BEYONDHD = "BeyondHD"
    PASSTHEPOPCORN = "PassThePopcorn"
    BROADCASTTHENET = "BroadcastTheNet"
    HDBITS = "HDBits"
    BLUTOPIA = "Blutopia"
    AITHER = "Aither"
    OTHER = "Other"
    OTHER_PRIVATE = "OtherPrivate"

    def is_public(self) -> bool:
        """
        Check if the current tracker is public.

        Returns
        -------
        bool
            `True` if the tracker is public, `False` otherwise.

        """
        return self.value in ("Nyaa", "AnimeTosho", "AniDex", "RuTracker", "Other")

    def is_private(self) -> bool:
        """
        Check if the current tracker is private.

        Returns
        -------
        bool
            `True` if the tracker is private, `False` otherwise.

        """
        return not self.is_public()

    @property
    def url(self) -> str:
        """
        URL of the current tracker.

        Returns
        -------
        str
            URL of the tracker.

        Notes
        -----
        Returns an empty string for [`Tracker.OTHER`][seadex.Tracker.OTHER]
        and [`Tracker.OTHER_PRIVATE`][seadex.Tracker.OTHER_PRIVATE].

        """
        _map = {
            "NYAA": b"aHR0cHM6Ly9ueWFhLnNp",
            "ANIMETOSHO": b"aHR0cHM6Ly9hbmltZXRvc2hvLm9yZw==",
            "ANIDEX": b"aHR0cHM6Ly9hbmlkZXguaW5mbw==",
            "RUTRACKER": b"aHR0cHM6Ly9ydXRyYWNrZXIub3Jn",
            "ANIMEBYTES": b"aHR0cHM6Ly9hbmltZWJ5dGVzLnR2",
            "BEYONDHD": b"aHR0cHM6Ly9iZXlvbmQtaGQubWU=",
            "PASSTHEPOPCORN": b"aHR0cHM6Ly9wYXNzdGhlcG9wY29ybi5tZQ==",
            "BROADCASTTHENET": b"aHR0cHM6Ly9icm9hZGNhc3RoZS5uZXQ=",
            "HDBITS": b"aHR0cHM6Ly9oZGJpdHMub3Jn",
            "BLUTOPIA": b"aHR0cHM6Ly9ibHV0b3BpYS5jYw==",
            "AITHER": b"aHR0cHM6Ly9haXRoZXIuY2M=",
            "OTHER": b"",
            "OTHER_PRIVATE": b"",
        }

        return base64.b64decode(_map[self.name]).decode()



from __future__ import annotations

import tomli

from seadex._version import __version__


def test_versions_match() -> None:
    with open("pyproject.toml", "rb") as f:
        pyproject_version = tomli.load(f)["project"]["version"]
    assert pyproject_version == __version__



from __future__ import annotations

try:
    from cyclopts import App
except ModuleNotFoundError:
    import sys

    print("Error: Required dependencies for the CLI are missing. Install `seadex[cli]` to fix this.")
    sys.exit(1)


from seadex._cli._backup import backup_app
from seadex._cli._entry import entry_app
from seadex._cli._torrent import torrent_app
from seadex._version import __version__

app = App(
    "seadex",
    version=__version__,
    help="Command line interface to the SeaDex API.",
    help_format="plaintext",
)

app.command(backup_app)
app.command(torrent_app)
app.command(entry_app)

if __name__ == "__main__":
    app()
